# Generated by Django 5.2.5 on 2025-10-15 15:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0004_book'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authorprofile',
            name='twitter_handle',
            field=models.CharField(blank=True, max_length=200),
        ),
    ]
